""" Fetchers are responsible for ingesting and standardizing data for future processing. """
import json
from typing import Any, Dict, List

import pandas as pd
import requests

import ftpvl.helpers as Helpers
from ftpvl.evaluation import Evaluation


class Fetcher:
    """
    This is a superclass for all fetchers.

    Fetchers allow the user to retrieve test data from a data source and output
    as an Evaluation for use by other tools in the library.
    """

    def __init__(self):
        self._eval_id = None

    def _download(self) -> Any:
        """
        Retrieves evaluation data over the internet and returns the data.
        """
        raise NotImplementedError

    def _preprocess(self, data: Any) -> pd.DataFrame:
        """
        Process the data and return the resulting dataframe.
        """
        raise NotImplementedError

    def get_evaluation(self) -> Evaluation:
        """
        Returns an Evaluation that represents the fetched data.
        """
        data = self._download()
        preprocessed_df = self._preprocess(data)
        return Evaluation(preprocessed_df, eval_id=self._eval_id)


class HydraFetcher(Fetcher):
    """
    Represents a downloader and preprocessor of test results from
    `hydra.vtr.tools`.

    Parameters
    ----------
    project : str, optional
        The project name to use when fetching from Hydra
    jobset : str, optional
        The jobset name to use when fetching from Hydra
    eval_num : int, optional
        A non-negative integer for the evaluation number to download,
        with `0` being the latest evaluation, by default 0
    mapping : dict, optional
        A dictionary mapping input column names to output
        column names, if needed for remapping, by default None
    hydra_clock_names : list, optional
        An optional ordered list of strings used in finding
        the actual frequency for each build result, by default None
    """

    def __init__(
        self,
        project: str,
        jobset: str,
        eval_num: int = 0,
        mapping: dict = None,
        hydra_clock_names: list = None
    ) -> None:
        super().__init__()
        self.project = project
        self.jobset = jobset
        self.eval_num = eval_num
        self.mapping = mapping
        self.hydra_clock_names = hydra_clock_names

    def _download(self) -> List[Dict]:
        """
        Fetches data from Hydra, returning a list of decoded meta.json dicts
        corresponding to the builds of the eval_num evaluation.

        Returns
        -------
        List[Dict]
            A list of dictionaries that represent the decoded meta.json files
            of each test result.

        Raises
        ------
        ConnectionError
            Raised if `hydra.vtr.tools` returns a non-200 status code when
            fetching evals.

        IndexError
            Raised if the specified eval_num is invalid due to it being too
            large.

        ValueError
            Raised if all builds in a given eval failed.
        """
        # get build numbers from eval_num
        resp = requests.get(
            f"https://hydra.vtr.tools/jobset/{self.project}/{self.jobset}/evals",
            headers={"Content-Type": "application/json"},
        )
        if resp.status_code != 200:
            raise ConnectionError("Unable to get evals from server.")
        evals_json = resp.json()
        if self.eval_num >= len(evals_json["evals"]):
            raise IndexError(f"Invalid eval_num: {self.eval_num}")
        build_nums = evals_json["evals"][self.eval_num]["builds"]

        # collect the 'meta.json' build products
        data = []
        for build_num in build_nums:
            resp = requests.get(
                f"https://hydra.vtr.tools/build/{build_num}/download/1/meta.json",
                headers={"Content-Type": "application/json"},
            )
            if resp.status_code != 200:
                print(
                    "Warning:",
                    f"Unable to get build {build_num}. It might have failed.",
                )
                continue
            try:
                data += [resp.json()]
            except json.decoder.JSONDecodeError:
                print("Warning:", f"Unable to decode build {build_num}")

        if len(data) == 0:
            raise ValueError(f"All builds from eval_num {self.eval_num} failed")

        self._eval_id = evals_json["evals"][self.eval_num]["id"]
        return data

    def _preprocess(self, data: List[Dict]) -> pd.DataFrame:
        """
        Using data from _download(), processes and standardizes the data and
        returns a Pandas DataFrame.
        """
        flattened_data = [Helpers.flatten(x) for x in data]

        processed_data = []
        for row in flattened_data:
            processed_row = {}
            if self.mapping is None:
                processed_row = row
            else:
                for in_col_name, out_col_name in self.mapping.items():
                    processed_row[out_col_name] = row[in_col_name]
            processed_row["freq"] = Helpers.get_actual_freq(row, self.hydra_clock_names)
            processed_row.update(Helpers.get_versions(row))
            processed_data.append(processed_row)

        return pd.DataFrame(processed_data).dropna(axis=1, how="all")
    
    def get_evaluation(self) -> Evaluation:
        return super().get_evaluation()


class JSONFetcher(Fetcher):
    """
    Represents a loader and preprocessor of test results from a JSON file.

    Parameters
    ----------
    path : str
        A string file path pointing to the dataframe JSON file.

    mapping : dict, optional
        An optional dictionary mapping input column names to output
        column names., by default None
    """

    def __init__(self, path: str, mapping: dict = None) -> None:

        super().__init__()
        self.path = path
        self.mapping = mapping

    def _download(self) -> str:
        """
        Returns the path of the JSON file.
        """
        return self.path

    def _preprocess(self, path: str) -> pd.DataFrame:
        """
        Given the path, load the data in pandas, process the data, and return
        the resulting dataframe.
        """
        loaded_df = pd.read_json(path)
        if self.mapping is None:
            return loaded_df
        else:
            return loaded_df.filter(items=self.mapping.keys()).rename(
                columns=self.mapping
            )

    def get_evaluation(self) -> Evaluation:
        return super().get_evaluation()



""" Styles are special processors that transform an evaluation into CSS styles. """
from matplotlib.colors import Colormap
from ftpvl.evaluation import Evaluation
from ftpvl.processors import Processor
from ftpvl.helpers import get_styling


class Style(Processor):
    """
    Represents a type of processor that returns an evaluation containing CSS
    styles for visualization.
    """

    def process(self, input_eval: Evaluation) -> Evaluation:
        raise NotImplementedError


class ColorMapStyle(Style):
    """
    Represents a processor that uses a matplotlib Colormap to create a styled
    evaluation where the background of a cell is specified by the value in the
    cell evaluated by the colormap.

    Values in the input are expected to be either a float between 0 and 1
    (inclusive) or a empty string.

    Parameters
    ----------
    cmap : Colormap
        the colormap to use when transforming input Evaluation to CSS styles.
    """

    def __init__(self, cmap: Colormap):
        self.cmap = cmap

    def process(self, input_eval: Evaluation) -> Evaluation:
        df = input_eval.get_df()
        new_df = df.applymap(lambda x: get_styling(x, self.cmap))
        return Evaluation(new_df)



""" Visualizers are used for displaying the results to the user in an IPython notebook. """
from typing import List

from ftpvl.evaluation import Evaluation

class Visualizer:
    """
    Superclass for visualizers that can generate and display styled evaluations.
    """

    def __init__(self):
        self._visualization = None

    def _generate(self):
        """
        Generates the visualzation and writes it to self._visualization
        """
        raise NotImplementedError

    def get_visualization(self):
        """
        Returns a displayable object which can be displayed by calling display()
        in an interactive Python environment.
        """
        self._generate()
        return self._visualization


class DebugVisualizer(Visualizer):
    """
    Represents a visualizer that will print the given Evaluation, possibly with
    version information.

    Useful for debugging.

    Parameters
    ----------
    evaluation : Evaluation
        the Evaluation to display

    version_info : bool, optional
        Flag to display version information from the build results in the
        final visualization, by default False

    custom_styles : List[dict], optional
        Specify additional styling for the final visualzation. See
        formatting here:
        https://pandas.pydata.org/pandas-docs/stable/user_guide/style.html#Table-styles,
        by default None

    column_order : List[str], optional
        Specify the columns and ordering  in the final visualization.
        Overrides version_info, so you must specify version columns in
        addition. Defaults to None, which will set the column order to a
        preset useful for VtR.

    Methods
    -------
    get_visualization() :
        returns a displayable visualizaton, can be displayed by calling
        display() in an interactive Python environment
    """

    def __init__(
        self,
        evaluation: Evaluation,
        version_info: bool = False,
        custom_styles: List[dict] = None,
        column_order: List[str] = None,
    ):
        super().__init__()
        self._evaluation = evaluation
        self._version_info = version_info
        self._custom_styles = custom_styles if custom_styles else []

        if column_order is None:
            self._column_order = []
            self._column_order += [
                "device",
                "bram",
                "carry",
                "dff",
                "iob",
                "lut",
                "pll",
                "synthesis",
                "pack",
                "place",
                "route",
                "fasm",
                "bitstream",
                "total",
                "freq",
                "normalized_max_freq",
            ]
            if self._version_info:
                self._column_order += [
                    "versions.vivado",
                    "versions.vpr",
                    "versions.yosys",
                    "versions.nextpnr-xilinx",
                    "versions.nextpnr-ice40",
                ]
        else:
            self._column_order = column_order

    def _generate(self):
        """ Generate visualization and save in self._visualization """
        ordered_df = self._evaluation.get_df()[self._column_order]

        self._visualization = (
            ordered_df.style.set_table_styles(self._custom_styles)
            .set_precision(2)
            .highlight_null("yellow")
            .set_na_rep("-")
        )
    
    def get_visualization(self):
        return super().get_visualization()


class SingleTableVisualizer(Visualizer):
    """
    Represents a visualizer for a styled single table, possibly with version
    information.

    Parameters
    ----------
    evaluation : Evaluation
        the Evaluation with values to display

    style_eval : Evaluation
        the Evaluation to use for styling. Should be processed using a
        Style, all values are valid CSS strings or empty.

    version_info : bool, optional
        Flag to display version information from the build results in the
        final visualization, by default False

    custom_styles : List[dict], optional
        Specify additional styling for the final visualzation. See
        formatting here:
        https://pandas.pydata.org/pandas-docs/stable/user_guide/style.html#Table-styles,
        by default None

    column_order : List[str], optional
        Specify the columns and ordering in the final visualization.
        Overrides version_info, so you must specify version columns in
        addition. Defaults to None, which will set the column order to a
        preset useful for VtR.

    Methods
    -------
    get_visualization() :
        returns a displayable visualizaton, can be
        displayed by calling display() in an interactive Python environment
    """

    def __init__(
        self,
        evaluation: Evaluation,
        style_eval: Evaluation,
        version_info: bool = False,
        custom_styles: List[dict] = None,
        column_order: List[str] = None,
    ):
        super().__init__()
        self._evaluation = evaluation
        self._style_eval = style_eval
        self._version_info = version_info

        self._custom_styles = custom_styles if custom_styles else []

        if column_order is None:
            self._column_order = []
            self._column_order += [
                "bram",
                "carry",
                "dff",
                "iob",
                "lut",
                "pll",
                "synthesis",
                "pack",
                "place",
                "route",
                "fasm",
                "bitstream",
                "total",
                "freq",
                "normalized_max_freq",
            ]
            if self._version_info:
                self._column_order += [
                    "versions.vivado",
                    "versions.vpr",
                    "versions.yosys",
                    "versions.nextpnr-xilinx",
                    "versions.nextpnr-ice40",
                ]
        else:
            self._column_order = column_order

    def _generate(self):
        """ Generate visualization and save in self._visualization """
        ordered_df = self._evaluation.get_df()[self._column_order]
        styled_df = self._style_eval.get_df()[self._column_order]
        self._visualization = (
            ordered_df.style.apply(lambda x: styled_df, axis=None)
            .set_table_styles(self._custom_styles)
            .set_precision(2)
            .highlight_null("yellow")
            .set_na_rep("-")
        )

    def get_visualization(self):
        return super().get_visualization()


# class TwoTableVisualizer(Visualizer):
#     """
#     Represents a visualizer for a styled double table.
#     """
#     def __init__(self, evaluation1: Evaluation, evaluation2: Evaluation,
#                  style: Style, version_info: bool = False):
#         super().__init__()
#         self._evaluation1 = evaluation1
#         self._evaluation2 = evaluation2
#         self._style = style
#         self._version_info = version_info

#     def _generate(self):
#         # TODO: Implement visualization generation
#         raise NotImplementedError



""" Defines helper functions useful for classes """
from typing import Union, Any

from matplotlib.colors import Colormap


def flatten(input_dict: dict) -> dict:
    """
    Given an input dictionary that may contain nested dictionaries, return
    a new dictionary that is flattened such that there are no dictionary
    values.

    Keys in the returned dictionary are a period-delimited list of
    the keys used to index the original dictionary.

    Parameters
    ----------
    input_dict : dict
        a potentially-nested dictionary

    Returns
    -------
    dict
        a new flattened input dictionary
    
    Examples
    --------
    >>> flatten({"a": {"b": "c"}})
    {"a.b": "c"}
    """
    new_dict = {}
    for k_1, v_1 in input_dict.items():
        if isinstance(v_1, dict):
            result = flatten(v_1)
            for k_2, v_2 in result.items():
                combined_key = f"{k_1}.{k_2}"
                new_dict[combined_key] = v_2
        else:
            new_dict[k_1] = v_1
    return new_dict


def get_versions(obj: dict) -> dict:
    """
    Given a flattened object decoded from meta.json, return a dictionary of
    the versions in that object.

    Filters the dictionary for all keys that start with `versions.`

    Parameters
    ----------
    obj : dict
        a flattened dictionary

    Returns
    -------
    dict
        a copy of the input dictionary only containing keys starting with
        `versions.`
    """
    return {k: v for k, v in obj.items() if k.startswith("versions.")}


def rescale_actual_freq(freq: Union[int, float]) -> Union[int, float]:
    """
    Given a frequency with an unspecified unit, returns frequency in megahertz
    by assuming the original unit is hertz if input frequency is greater than 1
    million.

    Parameters
    ----------
    freq : Union[int, float]
        a number that is a frequency

    Returns
    -------
    Union[int, float]
        the input frequency in megahertz
    """
    one_mhz = 1_000_000
    if freq > one_mhz:
        return freq / one_mhz
    else:
        return freq


def get_actual_freq(obj: dict, hydra_clock_names: list = None) -> Union[int, float]:
    """
    Given an object decoded from meta.json, return the actual frequency
    as an integer in megahertz.

    Since a meta.json object might contain multiple frequencies, we look through
    all clock names specified in hydra_clock_names and use the first one in the
    list. If none of the specified clock names exists in the object, we use
    the shortest clock name to find the frequency.

    Parameters
    ----------
    obj : dict
        A decoded meta.json file.

    hydra_clock_names : list, optional
        An ordered list of clock names to look for in the obj, by default None

    Returns
    -------
    Union[int, float]
        the frequency of the actual clock specified in the object
    """
    # set default clock names
    if hydra_clock_names is None:
        hydra_clock_names = ["clk", "sys_clk", "clk_i"]

    # if max_freq is unnested
    if "max_freq" in obj:
        return rescale_actual_freq(obj["max_freq"])
    else:
        # check if max_freq contains clock_name in HYDRA_CLOCK_NAMES
        for clock_name in hydra_clock_names:
            key = f"max_freq.{clock_name}.actual"
            if key in obj:
                return rescale_actual_freq(obj[key])

        # if none of those exist, choose the shortest one or return None
        max_freq_keys = [
            x for x in obj.keys() if x.startswith("max_freq.") and x.endswith(".actual")
        ]
        if len(max_freq_keys) > 0:
            shortest_clock_name = min(max_freq_keys, key=len)
            return rescale_actual_freq(obj[shortest_clock_name])
        else:
            return None


def get_styling(val: Any, cmap: Colormap) -> str:
    """
    Given a value, returns a CSS string with the background-color set to the
    color in the cmap, or an empty CSS string if the value is not a float
    between 0 and 1.

    Parameters
    ----------
    val : Any
        A value that needs to be styled

    cmap : Colormap
        A colormap to use when styling the background

    Returns
    -------
    str
        a CSS string specifying the background color based on the color map,
        or an empty string if the val is not a float between 0 and 1.
    """
    if isinstance(val, float) and 0 <= val <= 1:
        color = tuple([int(x * 255) for x in cmap(val)[:-1]])
        hex_color = "#%02x%02x%02x" % color  # convert to hex format #FFFFFF
        return "background-color: {}".format(hex_color)
    else:
        return ""
